@@ -3,11 +3,12 @@ import time
 from collections import namedtuple
 from logging import getLogger
 from pathlib import Path
+from tempfile import TemporaryDirectory
 from typing import Dict, List, Optional, Union
 from urllib.parse import urljoin, urlparse, urlunparse
 
-from .auth import get_token
-from .exceptions import ClientValueError
+from .auth import Authentication
+from .exceptions import ClientException, ClientValueError, ExtraPackageError
 from .response import Result
 from .session import HttpSession
 from .utils import create_md5, get_media_type
@@ -67,14 +68,6 @@ class Client(object):
             )
         self.session = session
 
-        if client_id is None or client_id == "":
-            raise ClientValueError(f"Client ID is not set.")
-        self.client_id = client_id
-
-        if client_secret is None or client_secret == "":
-            raise ClientValueError(f"Client Secret is not set.")
-        self.client_secret = client_secret
-
         parts = urlparse(api_url)
         api_path = parts.path
         if "anymotion" not in api_path:
@@ -82,27 +75,18 @@ class Client(object):
         if api_path[-1] != "/":
             api_path += "/"
 
-        self._base_url = str(urlunparse((parts.scheme, parts.netloc, "", "", "", "")))
-        self._api_url = urljoin(self._base_url, api_path)
-        self._token: Optional[str] = None
+        base_url = str(urlunparse((parts.scheme, parts.netloc, "", "", "", "")))
+        self._api_url = urljoin(base_url, api_path)
+
+        self.auth = Authentication(
+            client_id, client_secret, base_url=base_url, session=self.session
+        )
 
         self._interval = max(0.1, interval)
         self._max_steps = int(max(1, timeout / self._interval))
 
         self._page_size = 1000
 
-    @property
-    def token(self) -> str:
-        """Return access token."""
-        if self._token is None:
-            self._token = get_token(
-                self.client_id,
-                self.client_secret,
-                base_url=self._base_url,
-                session=self.session,
-            )
-        return self._token
-
     def get_one_data(self, endpoint: str, endpoint_id: int) -> dict:
         """Get one piece of data.
 
@@ -117,7 +101,7 @@ class Client(object):
             RequestsError: HTTP request fails.
         """
         url = urljoin(self._api_url, f"{endpoint}/{endpoint_id}/")
-        response = self.session.request(url, token=self.token)
+        response = self.session.request(url, token=self.auth.token)
         return response.json
 
     def get_image(self, image_id: int) -> dict:
@@ -128,17 +112,35 @@ class Client(object):
         """Get movie data."""
         return self.get_one_data("movies", movie_id)
 
-    def get_keypoint(self, keypoint_id: int) -> dict:
+    def get_keypoint(self, keypoint_id: int, join_data: bool = False) -> dict:
         """Get keypoint data."""
-        return self.get_one_data("keypoints", keypoint_id)
+        keypoint = self.get_one_data("keypoints", keypoint_id)
+        if join_data:
+            image_id = keypoint.get("image")
+            if image_id:
+                keypoint["image"] = self.get_image(image_id)
+            movie_id = keypoint.get("movie")
+            if movie_id:
+                keypoint["movie"] = self.get_movie(movie_id)
+        return keypoint
 
-    def get_drawing(self, drawing_id: int) -> dict:
+    def get_drawing(self, drawing_id: int, join_data: bool = False) -> dict:
         """Get drawing data."""
-        return self.get_one_data("drawings", drawing_id)
+        drawing = self.get_one_data("drawings", drawing_id)
+        if join_data:
+            keypoint_id = drawing.get("keypoint")
+            if keypoint_id:
+                drawing["keypoint"] = self.get_keypoint(keypoint_id, join_data=True)
+        return drawing
 
-    def get_analysis(self, analysis_id: int) -> dict:
+    def get_analysis(self, analysis_id: int, join_data: bool = False) -> dict:
         """Get analysis data."""
-        return self.get_one_data("analyses", analysis_id)
+        analysis = self.get_one_data("analyses", analysis_id)
+        if join_data:
+            keypoint_id = analysis.get("keypoint")
+            if keypoint_id:
+                analysis["keypoint"] = self.get_keypoint(keypoint_id, join_data=True)
+        return analysis
 
     def get_list_data(self, endpoint: str, params: dict = {}) -> List[dict]:
         """Get list data.
@@ -150,7 +152,7 @@ class Client(object):
         params["size"] = self._page_size
         data: List[dict] = []
         while url:
-            response = self.session.request(url, params=params, token=self.token)
+            response = self.session.request(url, params=params, token=self.auth.token)
             sub_data, url = response.get(("data", "next"))
             params = {}
             data += sub_data
@@ -161,15 +163,15 @@ class Client(object):
         return self.get_list_data("images", params=params)
 
     def get_movies(self, params: dict = {}) -> List[dict]:
-        """Get movie data."""
+        """Get movie list."""
         return self.get_list_data("movies", params=params)
 
     def get_keypoints(self, params: dict = {}) -> List[dict]:
-        """Get keypoint data."""
+        """Get keypoint list."""
         return self.get_list_data("keypoints", params=params)
 
     def get_drawings(self, params: dict = {}) -> List[dict]:
-        """Get drawing data."""
+        """Get drawing list."""
         return self.get_list_data("drawings", params=params)
 
     def get_analyses(self, params: dict = {}) -> List[dict]:
@@ -206,7 +208,7 @@ class Client(object):
             urljoin(self._api_url, f"{media_type}s/"),
             method="POST",
             json={"content_md5": content_md5, "name": path.stem, "text": text},
-            token=self.token,
+            token=self.auth.token,
         )
         media_id, upload_url = response.get(("id", "uploadUrl"))
 
@@ -229,8 +231,8 @@ class Client(object):
         path: Optional[Union[str, Path]] = None,
         exist_ok: bool = False,
         fix_suffix: bool = False,
-    ) -> None:
-        """Download file from drawing_id.
+    ) -> Path:
+        """Download a file from drawing_id.
 
         Args:
             drawing_id
@@ -239,15 +241,20 @@ class Client(object):
                 already exists.
             fix_suffix: If the extension of path is invalid, correct it.
 
+        Returns:
+            The path to the downloaded file.
+
         Raises:
+            ClientException
             FileExistsError
             RequestsError: HTTP request fails.
         """
         data = self.get_one_data("drawings", drawing_id)
         url = data.get("drawingUrl")
         if url is None:
-            logger.warning("Skip download because there is no drawing url.")
-            return
+            raise ClientException(
+                "Can't download the file because it doesn't have a drawing url."
+            )
         url_path = Path(urlparse(url).path)
 
         if path is None:
@@ -273,6 +280,25 @@ class Client(object):
             f.write(response.raw.content)
         logger.info(f"Download file to {path}.")
 
+        return path
+
+    def download_and_read(self, drawing_id: int):
+        """Download and read a file from drawing_id."""
+        try:
+            from .extras import read_image, read_video
+        except ImportError:
+            raise ExtraPackageError(
+                "The extras package is not installed. "
+                "Install as follows: pip install anymotion-sdk[cv]"
+            )
+
+        with TemporaryDirectory() as dir_path:
+            file_path = self.download(drawing_id, path=dir_path)
+            if get_media_type(file_path) == "image":
+                return read_image(file_path)
+            else:
+                return read_video(file_path)
+
     def extract_keypoint(
         self,
         data: Optional[dict] = None,
@@ -305,7 +331,9 @@ class Client(object):
             data = {"image_id": image_id}
 
         url = urljoin(self._api_url, "keypoints/")
-        response = self.session.request(url, method="POST", json=data, token=self.token)
+        response = self.session.request(
+            url, method="POST", json=data, token=self.auth.token
+        )
         return response.get("id")
 
     def draw_keypoint(
@@ -319,11 +347,13 @@ class Client(object):
         Raises:
             RequestsError: HTTP request fails.
         """
-        url = urljoin(self._api_url, f"drawings/")
+        url = urljoin(self._api_url, "drawings/")
         json: Dict[str, Union[int, list, dict]] = {"keypoint_id": keypoint_id}
         if rule is not None:
             json["rule"] = rule
-        response = self.session.request(url, method="POST", json=json, token=self.token)
+        response = self.session.request(
+            url, method="POST", json=json, token=self.auth.token
+        )
         return response.get("id")
 
     def analyze_keypoint(self, keypoint_id: int, rule: Union[list, dict]) -> int:
@@ -335,11 +365,13 @@ class Client(object):
         Raises:
             RequestsError: HTTP request fails.
         """
-        url = urljoin(self._api_url, f"analyses/")
+        url = urljoin(self._api_url, "analyses/")
         json: Dict[str, Union[int, list, dict]] = {"keypoint_id": keypoint_id}
         if rule is not None:
             json["rule"] = rule
-        response = self.session.request(url, method="POST", json=json, token=self.token)
+        response = self.session.request(
+            url, method="POST", json=json, token=self.auth.token
+        )
         return response.get("id")
 
     def wait_for_extraction(self, keypoint_id: int) -> Result:
@@ -371,7 +403,7 @@ class Client(object):
 
     def _wait_for_done(self, url: str) -> Result:
         for _ in range(self._max_steps):
-            response = self.session.request(url, token=self.token)
+            response = self.session.request(url, token=self.auth.token)
             result = Result(response.raw)
             if result.status in ["SUCCESS", "FAILURE"]:
                 break
